/*
 * This file and its contents are licensed under the Apache License 2.0.
 * Please see the included NOTICE for copyright information and
 * LICENSE-APACHE for a copy of the license.
 */
#pragma once

/*
 * This file defines the node name of Vector Aggregation custom node, to be
 * used in the Apache part of the Timescale extension. The node itself is in the
 * the TSL part.
 */
#define VECTOR_AGG_NODE_NAME "VectorAgg"
