-- This file and its contents are licensed under the Apache License 2.0.
-- Please see the included NOTICE for copyright information and
-- LICENSE-APACHE for a copy of the license.

CREATE DATABASE single;

\c single
CREATE EXTENSION IF NOT EXISTS timescaledb;

