/*
 * This file and its contents are licensed under the Apache License 2.0.
 * Please see the included NOTICE for copyright information and
 * LICENSE-APACHE for a copy of the license.
 */
/* Overwrite main config.h so we could inject our own version #s here. */
#define TELEMETRY_DEFAULT TELEMETRY_OFF
