# This file and its contents are licensed under the Timescale License.
# Please see the included NOTICE for copyright information and
# LICENSE-TIMESCALE for a copy of the license.

package DataNode;
use parent qw(TimescaleNode);
use strict;
use warnings;

1;
