/*
 * This file and its contents are licensed under the Apache License 2.0.
 * Please see the included NOTICE for copyright information and
 * LICENSE-APACHE for a copy of the license.
 */
#pragma once

#include <postgres.h>
#include <nodes/extensible.h>

#include "export.h"

extern TSDLLEXPORT Expr *ts_transform_cross_datatype_comparison(Expr *clause);
