/*
 * This file and its contents are licensed under the Apache License 2.0.
 * Please see the included NOTICE for copyright information and
 * LICENSE-APACHE for a copy of the license.
 */

#pragma once

#include "export.h"

extern TSDLLEXPORT bool ts_is_valid_timezone_name(const char *tz_name);
