/*
 * This file and its contents are licensed under the Apache License 2.0.
 * Please see the included NOTICE for copyright information and
 * LICENSE-APACHE for a copy of the license.
 */
#include "extension_constants.h"

const char *const ts_extension_schema_names[] = {
	[TS_CATALOG_SCHEMA] = CATALOG_SCHEMA_NAME,
	[TS_FUNCTIONS_SCHEMA] = FUNCTIONS_SCHEMA_NAME,
	[TS_INTERNAL_SCHEMA] = INTERNAL_SCHEMA_NAME,
	[TS_CACHE_SCHEMA] = CACHE_SCHEMA_NAME,
	[TS_CONFIG_SCHEMA] = CONFIG_SCHEMA_NAME,
	[TS_EXPERIMENTAL_SCHEMA] = EXPERIMENTAL_SCHEMA_NAME,
	[TS_INFORMATION_SCHEMA] = INFORMATION_SCHEMA_NAME,
};
