-- This file and its contents are licensed under the Apache License 2.0.
-- Please see the included NOTICE for copyright information and
-- LICENSE-APACHE for a copy of the license.

\ir cleanup.bigint.sql
\ir cleanup.constraints.sql
\ir cleanup.timestamp.sql
\ir cleanup.continuous_aggs.v2.sql
\ir cleanup.compression.sql
\ir cleanup.policies.sql

